@@ -106,9 +106,6 @@ from linref.route import MLSRoute, combine_mpgs
 from rangel import RangeCollection
 import copy, warnings
 
-# Temporary dependencies
-from linref.various.geospatial import join_nearby
-
 
 ##################
 # EVENTS CLASSES #
@@ -187,14 +184,35 @@ class EventsFrame(object):
             if isinstance(df, gpd.GeoDataFrame) and self.geom is None:
                 self.geom = df.geometry.name
 
-            # Reset log
+            # Reset logs
             try:
                 self.log.reset()
             except:
                 pass
+            self._initialize_df()
         else:
-            raise TypeError("Input dataframe must be pandas DataFrame class "
-                "instance.")
+            raise TypeError(
+                "Input dataframe must be pandas DataFrame class instance.")
+
+    @property
+    def size(self):
+        """
+        Return the size of the events dataframe.
+        """
+        return self._df.size
+
+    @property
+    def shape(self):
+        """
+        Return the shape of the events dataframe.
+        """
+        return self._df.shape
+
+    def _initialize_df(self):
+        """
+        Class-specific dataframe initialization processes.
+        """
+        pass
 
     def _sort_df(self, df):
         """
@@ -204,6 +222,30 @@ class EventsFrame(object):
         return df.sort_values(
             by=self.keys + [self.beg, self.end], ascending=True)
 
+    def set_df(self, obj, inplace=False):
+        """
+        Set a new events dataframe.
+        """
+        # Define target, copy if needed
+        ef = self if inplace else self.copy()
+        # Assign dataframe
+        ef.df = obj
+        # Return if needed
+        if not inplace:
+            return ef
+
+    def sort(self, inplace=False):
+        """
+        Sort the events dataframe based on target columns.
+        """
+        # Create a copy if requested
+        ef = self if inplace else self.copy()
+        # Log sorting
+        ef._sort = True
+        ef.df = self.df
+        if not inplace:
+            return ef
+
     def df_exportable(self):
         """
         Return a dataframe which is optimized for exporting.
@@ -278,7 +320,7 @@ class EventsFrame(object):
         """
         A list of all columns within the events dataframe.
         """
-        return list(self._df.columns)
+        return self._df.columns.values.tolist()
 
     @property
     def targets(self):
@@ -827,11 +869,11 @@ class EventsFrame(object):
         ec = EventsCollection(res, keys=self.keys, beg=self.beg, end=self.end, 
             geom=self.geom if agg_geometry else None,
             route='route' if agg_routes else None, 
-            closed=self.closed)
+            closed=self.closed, missing_data='ignore')
         return ec
 
     def project(self, other, buffer=100, nearest=True, loc_label='LOC', 
-            dist_label='DISTANCE', **kwargs):
+            dist_label='DISTANCE', build_routes=True, **kwargs):
         """
         Project an input geodataframe onto the events dataframe, producing 
         linearly referenced point locations relative to events for all input 
@@ -848,11 +890,15 @@ class EventsFrame(object):
             coordinate reference system.
         nearest : bool, default True
             Whether to choose only the nearest match within the defined buffer. 
-            If False, all matches will be returned.
+            If False, all matches will be returned. If True, when multiple 
+            equidistant points exist, choose the first result that appears.
         loc_label, dist_label : label
             Labels to be used for created columns for projected locations on 
             target events groups and nearest point distances between target 
             geometries and events geometries.
+        build_routes : bool, default True
+            Whether to automatically build routes using the build_routes() 
+            method if routes are not already available.
         **kwargs
             Keyword arguments to be passed to the EventsFrame constructor 
             upon completion of the projection.
@@ -860,6 +906,12 @@ class EventsFrame(object):
         # Validate input geodataframe
         if not isinstance(other, gpd.GeoDataFrame):
             raise TypeError("Other object must be gpd.GeoDataFrame instance.")
+        else:
+            try:
+                other_geometry = other.geometry.name
+            except AttributeError:
+                raise AttributeError(
+                    "No geometry data set in other geodataframe.")
         other = other.copy()
 
         # Check for invalid column names
@@ -875,54 +927,42 @@ class EventsFrame(object):
 
         # Ensure that geometries and routes are available
         if self.geom is None:
-            raise ValueError("No geometry found in events dataframe. If "
-                "valid shapely geometries are available in the dataframe, "
-                f"set this with the {self.__class__.__name__}'s geom "
-                "property.")
+            raise ValueError(
+                "No geometry found in events dataframe. If valid shapely "
+                "geometries are available in the dataframe, set this with the "
+                f"{self.__class__.__name__}'s geom property.")
         elif self.route is None:
-            raise ValueError("No routes found in events dataframe. If valid "
-                "shapely geometries are available in the dataframe, create "
-                "routes by calling the build_routes() method on the "
-                f"{self.__class__.__name__} class instance.")
+            if build_routes:
+                self.build_routes()
+            else:
+                raise ValueError(
+                    "No routes found in events dataframe. If valid shapely "
+                    "geometries are available in the dataframe, create routes "
+                    "by calling the build_routes() method on the "
+                    f"{self.__class__.__name__} class instance.")
         
         # Join the other geodataframe to this one
         select_cols = self.keys + [self.route, self.geom]
-        try:
-            if nearest:
-                joined = other.sjoin_nearest(
-                    self.df[select_cols],
-                    max_distance=buffer,
-                    how='left'
-                )
-            else:
-                warnings.warn(
-                    "Performance when nearest=False is currently limited and "
-                    "will be improved in future versions.")
-                joined = join_nearby(
-                    other, 
-                    self.df[select_cols], 
-                    buffer=buffer, 
-                    choose='all',
-                    dist_label=dist_label
-                )
-        except AttributeError:
-            # Optional dependency warning for improved performance
-            warnings.warn(
-                "Performance will be reduced for this operation when using "
-                "the current geopandas version. Upgrade to geopandas>=v0.10.0 "
-                "for improved performance.")
-            joined = join_nearby(
-                other, 
-                self.df[select_cols], 
-                buffer=buffer, 
-                choose='min' if nearest else 'all',
-                dist_label=dist_label
+        if nearest:
+            joined = other.sjoin_nearest(
+                self.df[select_cols],
+                max_distance=buffer,
+                how='left'
+            )
+            # Drop duplicates (required for equidistant ties)
+            joined = joined[~joined.index.duplicated(keep='first')]
+        else:
+            # Buffer geometry for spatial join
+            buffered_geoms = self.df.geometry.buffer(buffer)
+            joined = other.sjoin(
+                self.df[select_cols].set_geometry(buffered_geoms),
+                how='left'
             )
 
         # Project input geometries onto event geometries
         def _project(r):
             try:
-                return r[self.route].project(r.geometry)
+                return r[self.route].project(r[other_geometry])
             except AttributeError:
                 return
         locs = joined.apply(_project, axis=1)
@@ -934,91 +974,7 @@ class EventsFrame(object):
             keys=self.keys,
             beg=loc_label,
             closed=self.closed,
-            **kwargs
-        )
-
-    def project_old(self, other, buffer=100, choose='min', loc_label='LOC', 
-            dist_label='DISTANCE', **kwargs):
-        """
-        Project an input geodataframe onto the events dataframe, producing 
-        linearly referenced point locations relative to events for all input 
-        geometries within a buffered search area.
-
-        Parameters
-        ----------
-        other : gpd.GeoDataFrame
-            Geodataframe containing geometry which will be projected onto the 
-            events dataframe.
-        buffer : float, default 100
-            The max distance to search for input geometries to project against 
-            the events' geometries. Measured in terms of the geometries' 
-            coordinate reference system.
-        choose : {'min', 'max', 'all'}, default 'min'
-            Which target geometry to choose when more than one falls within the 
-            buffer distance.
-
-            Options
-            -------
-            min : choose the geometry with the shortest distance from the 
-                events data
-            max : choose the geometry with the longest distance from the 
-                events data
-            all : return all geometries which fall within the buffer area
-        loc_label, dist_label : label
-            Labels to be used for created columns for projected locations on 
-            target events groups and nearest point distances between target 
-            geometries and events geometries.
-        **kwargs
-            Keyword arguments to be passed to the EventsFrame constructor 
-            upon completion of the projection.
-        """
-        # Validate input geodataframe
-        if not isinstance(other, gpd.GeoDataFrame):
-            raise TypeError("Other object must be gpd.GeoDataFrame instance.")
-        other = other.copy()
-
-        # Check for invalid column names
-        if self.route in other.columns:
-            raise ValueError(f"Invalid column name '{self.route}' found in "
-                "target geodataframe.")
-
-        # Ensure that geometries and routes are available
-        if self.geom is None:
-            raise ValueError("No geometry found in events dataframe. If "
-                "valid shapely geometries are available in the dataframe, "
-                f"set this with the {self.__class__.__name__}'s geom "
-                "property.")
-        elif self.route is None:
-            raise ValueError("No routes found in events dataframe. If valid "
-                "shapely geometries are available in the dataframe, create "
-                "routes by calling the build_routes() method on the "
-                f"{self.__class__.__name__} class instance.")
-        
-        # Join the other geodataframe to this one
-        select_cols = self.keys + [self.route, self.geom]
-        joined = join_nearby(
-            other, 
-            self.df[select_cols], 
-            buffer=buffer, 
-            choose=choose,
-            dist_label=dist_label
-        )
-
-        # Project input geometries onto event geometries
-        def _project(r):
-            try:
-                return r[self.route].project(r.geometry)
-            except AttributeError:
-                return
-        locs = joined.apply(_project, axis=1)
-        joined[loc_label] = locs
-
-        # Prepare and return data
-        return self.__class__(
-            joined.drop(columns=[self.route]),
-            keys=self.keys,
-            beg=loc_label,
-            closed=self.closed,
+            missing_data='ignore',
             **kwargs
         )
 
@@ -1096,7 +1052,13 @@ class EventsFrame(object):
         }
         dtypes = {col: dtypes[col] for col in df.columns}
         df = df.astype(dtypes, copy=False)
-        res = self.__class__(df, keys=self.keys, beg=self.beg, end=self.end)
+        res = self.__class__(
+            df,
+            keys=self.keys,
+            beg=self.beg,
+            end=self.end,
+            missing_data='ignore'
+        )
         return res
 
     def to_windows(self, dissolve=False, endpoint=False, **kwargs):
@@ -1182,7 +1144,13 @@ class EventsFrame(object):
         }
         dtypes = {col: dtypes[col] for col in df.columns}
         df = df.astype(dtypes, copy=False)
-        res = self.__class__(df, keys=self.keys, beg=self.beg, end=self.end)
+        res = self.__class__(
+            df,
+            keys=self.keys,
+            beg=self.beg,
+            end=self.end,
+            missing_data='ignore'
+        )
         return res
 
 
@@ -1732,10 +1700,21 @@ class EventsCollection(EventsFrame):
     sort : bool, default False
         Whether to sort the events dataframe by its keys and begin and end 
         values upon its creation.
+    missing_data : {'ignore','drop','warn','raise'}, default 'warn'
+        What to do when the input dataframe contains missing values in the 
+        target key, beg, and end columns.
+
+        Options
+        -------
+        ignore : do nothing.
+        drop : drop all records which contain any missing data in the target 
+            columns.
+        warn : log a warning when records are missing data.
+        raise : raise a ValueError when records are missing data.
     """
 
     def __init__(self, df, keys=None, beg=None, end=None, 
-        geom=None, closed=None, sort=False, **kwargs):
+        geom=None, closed=None, sort=False, missing_data='warn', **kwargs):
         # Validate keys option
         if keys is None:
             raise Exception("If no keys are required to define unique groups "
@@ -1747,6 +1726,9 @@ class EventsCollection(EventsFrame):
         # Log input values
         self.closed = closed
         
+        # Additional processing
+        self._check_missing_data(missing_data=missing_data)
+
         # Create events log
         self.log = EventsLog()
         
@@ -1767,6 +1749,48 @@ class EventsCollection(EventsFrame):
                 # Single group
                 return self.get_group(keys, empty=False)
 
+    def _initialize_df(self):
+        """
+        Class-specific dataframe initialization processes.
+        """
+        self._empty_df = pd.DataFrame(columns=self.columns)
+        self._empty_group = self._build_group(self._empty_df.copy())
+
+    def _check_missing_data(self, missing_data='warn'):
+        """
+        Check for missing data in keys, beg, end, and geometry fields. Warn 
+        user when target fields contain null data.
+        """
+        # If ignore
+        if missing_data=='ignore':
+            return
+        elif missing_data in ['warn','raise','drop']:
+            # Find, count missing data records
+            mask = self.df[self.targets].isna().any(axis=1)
+            count = mask.sum()
+            # Address if more than one records contain missing data
+            if count > 0:
+                # Drop records
+                if missing_data=='drop':
+                    self.df = self.df[~mask].copy()
+                    return
+                # Warn or raise error
+                else:
+                    # Prepare message
+                    message = (
+                        f"Input events dataframe has {count:,.0f} records "
+                        "with missing data in target columns. This may cause "
+                        "unexpected behaviors.")
+                    if missing_data=='raise':
+                        raise ValueError(message)
+                    else:
+                        warnings.warn(message)
+                        return
+        else:
+            raise ValueError(
+                "Invalid input missing_data parameter. Must be one of "
+                "('ignore','drop','warn','raise').")
+
     def from_similar(self, df, **kwargs):
         """
         Create an EventsCollection from the input dataframe, assuming the same 
@@ -1880,7 +1904,7 @@ class EventsCollection(EventsFrame):
         self._log = obj
     
     def _build_empty(self):
-        return pd.DataFrame(columns=self.df.columns)
+        return self._empty_group.copy(deep=True)
     
     def reset_log(self):
         """
@@ -1897,8 +1921,8 @@ class EventsCollection(EventsFrame):
         # Validate input keys
         if self.num_keys == 0:
             if not keys is None:
-                raise ValueError("No keys defined in the collection to be "
-                    "queried.")
+                raise ValueError(
+                    "No keys defined in the collection to be queried.")
         elif self.num_keys == 1:
             if isinstance(keys, list) or isinstance(keys, tuple):
                 keys = keys[0]
@@ -2018,10 +2042,12 @@ class EventsCollection(EventsFrame):
             beg=self.beg,
             end=self.end,
             closed=self.closed,
+            missing_data='ignore',
             **kwargs
         )
     
-    def get_group(self, keys, empty=True, **kwargs) -> EventsGroup:
+    def get_group(self, keys, empty=True, log_empty=True, 
+            **kwargs) -> EventsGroup:
         """
         Retrieve a unique group of events based on provided key values.
  
@@ -2030,60 +2056,42 @@ class EventsCollection(EventsFrame):
         keys : key value, tuple of key values, or list of the same
             If only one key column is defined within the collection, a single 
             column value may be provided. Otherwise, a tuple of column values 
-            must be provided in the same order as they appear in self.keys. To 
-            get multiple groups, a list of key values or tuples may be 
-            provided.
+            must be provided in the same order as they appear in self.keys.
         empty : bool, default True
             Whether to allow for empty events groups to be returned when the 
             provided keys are valid but are not associated with any actual 
             events. If False, these cases will return a KeyError.
+        log_empty : bool, default True
+            Whether created empty events should be logged and stored within 
+            the collection to allow for quicker access. More memory intensive 
+            but may produce moderate performance improvements if empty keys 
+            will be accessed repeatedly.
         """
-        # Enforce multiple keys method
-        if not isinstance(keys, list):
-            keys = [keys]
-            ndim = 1
-        else:
-            ndim = len(keys)
-
-        # Iterate over keys
-        dfs = []
-        for keys_i in keys:
-            # Attempt to retrieve from log
-            keys_i = self._validate_keys(keys_i)
+        # Attempt to retrieve from log
+        keys = self._validate_keys(keys)
+        try:
+            # Retrieve from log
+            group = self.log[keys]
+        except KeyError:
+            # Attempt to retrieve dataframe to create new group
             try:
-                dfs.append(self.log[keys_i].df)
+                # Build and add group to log
+                group = self._build_group(self._groups.get_group(keys))
+                self.log[keys] = group
+            # Invalid group keys (i.e., empty group)
             except KeyError:
-                # Attempt to retrieve group
-                try:
-                    df = self.groups.get_group(keys_i)
-                    dfs.append(df)
-                    # Add group to log
-                    self.log[keys_i] = self._build_group(df)
-                # - Collection is None (i.e., no defined keys)
-                except AttributeError:
-                    dfs.append(self.df)
-                    break
-                # - Invalid group keys (i.e., empty group)
-                except KeyError as e:
-                    # Deal with empty group
-                    if empty:
-                        dfs.append(self._build_empty())
-                    else:
-                        raise KeyError(
-                            f"Invalid EventsCollection keys: {keys_i}")
-        # Log and return retrieved group
-        if ndim == 1:
-            return self._build_group(dfs[0])
-        else:
-            df = pd.concat(dfs)
-            try:
-                df = gpd.GeoDataFrame(
-                    df, geometry=self.geom, crs=self.df.crs)
-            except:
-                pass
-            return EventsCollection(
-                df=df, keys=self.keys, beg=self.beg, end=self.end, 
-                geom=self.geom, closed=self.closed)
+                # Deal with empty group
+                if empty:
+                    group = self._build_empty()
+                    if log_empty:
+                        self.log[keys] = group
+                else:
+                    raise KeyError(
+                        f"Invalid EventsCollection keys: {keys}")
+            # Collection is None (i.e., no defined keys)
+            except AttributeError:
+                raise ValueError("No defined group keys.")
+        return group
 
     def get_subset(self, keys, reduce=True, **kwargs):
         """
@@ -2133,10 +2141,11 @@ class EventsCollection(EventsFrame):
         try:
             ec = EventsCollection(
                 df, keys=new_keys, beg=self.beg, end=self.end, 
-                geom=self.geom, closed=self.closed)
+                geom=self.geom, closed=self.closed, missing_data='ignore')
         except:
-            display(df)
-            raise ValueError()
+            raise ValueError(
+                "Unable to produce EventsCollection subset due to unknown "
+                "error.")
         return ec
 
     def get_matching(self, other, **kwargs):
@@ -2163,15 +2172,54 @@ class EventsCollection(EventsFrame):
             return EventsGroup(
                 df=df, beg=self.beg, end=self.end, geom=self.geom, 
                 closed=self.closed)
-        except:
+        except Exception as e:
             display(df)
-            raise ValueError()
+            raise e
 
 
 ####################
 # COMMON FUNCTIONS #
 ####################
 
+def from_standard(df, require_end=False, **kwargs):
+    """
+    Create an EventsCollection from the input dataframe assuming standard 
+    column labels. These standard labels can be modified on the class 
+    directly be modifying the associated class attributes:
+    - default_keys
+    - default_beg
+    - default_end
+    - default_geom
+
+    Standard labels include:
+    keys : 'RID', 'YEAR', 'KEY'
+    beg : 'BMP', 'BEG', 'FROM'
+    end : 'EMP', 'END', 'TO'
+    geom : 'geometry'
+
+    Additional constructor keyword arguments can be passed through 
+    **kwargs.
+
+    Parameters
+    ----------
+    df : pd.DataFrame
+        Pandas dataframe which contains linear or point events data, 
+        formatted with standard labels. If multiple keys are detected, they 
+        will be assigned in the order in which they appear within the 
+        target dataframe. Only one of each begin and end option may be 
+        used. The geometry label is optional.
+    require_end : bool, default False
+        Whether to raise an error if no valid unique end column label is 
+        found. If False, no end label will be used when generating the 
+        collection.
+    **kwargs
+        Additional keyword arguments to be passed to the EventsCollection 
+        constructor.
+    """
+    ec = EventsCollection.from_standard(df, require_end=require_end, **kwargs)
+    return ec
+
+
 def check_compatibility(objs, errors='raise', **kwargs):
     """
     Check if the input list of EventsCollections are all compatible for 
